==> lmtpd-0.9.19.tar.gz: lmtpd-0.9.19/chk.c <==
/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

static char const rcsid[] =
"$Id: chk.c,v 1.29 2004/01/28 15:56:45 lwa Exp $";

#include "config.h"

#include <sys/types.h>
#include <regex.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "lmtp.h"

#define TRUE  1
#define FALSE 0

#if 0
#include <stdio.h> 
#define DEBUG(x...) printf(##x)
#else
#define DEBUG(x...)
#endif

static char *utf8_header(struct headers *h) {
  if (h->utf8)
    return h->utf8;
  switch(h->type) {
  case HEADER_DELIVERED_TO:
  case HEADER_RETURN_PATH:
  case HEADER_MESSAGE_ID:
  case HEADER_RECEIVED:
    h->utf8 = h->header;
  default:
    h->utf8 = rfc2047_read(h->header, h->header+h->dd_len);
  }
  return h->utf8;
}

/*
 * return 0 if RE is valid
 */

int valid_re(RE *re) {
  switch(re->err) {
  case RE_ERR_MEMORY:
    errno = ENOMEM;
    return -1;
    break;
  case RE_ERR_SYNTAX:
    errno = EINVAL;
    return -1;
    break;
  }
  return 0;
}

/*
 *  extract first hdr header from hdrlist
 */

char *tochk(char **hdr, struct headers *hdrlist) {
  char *ret;
  int convert_utf8 = 0;

  int (*cmp)(const char *, const char *, size_t);

  if (hdr) {
    char *xhdr=*hdr;
    char *chdr=*hdr;
    uint32_t set = 0;
    convert_utf8 = 0;

    if (*xhdr=='\\') {
      xhdr++; chdr++;
      cmp = strncmp;
    } else {
      if (*xhdr == '~') {
	xhdr++;
	chdr++;
	convert_utf8 = 1;
      }
      if (*xhdr == '%')
	set = get_set(xhdr[1]);
      cmp = strncasecmp;
    }
    while(1) {
      if (*xhdr == '\0' || *xhdr == ' ' || *xhdr == '\t') {
	*hdr=NULL;
	return NULL;
      }
      if (*xhdr==':')
	break;
      xhdr++;
    }
    *hdr=xhdr;
    while (**hdr==':') (*hdr)++;

    DEBUG("chdr=%s, xhdr=%s, *hdr=%s, dd_len=%d, xhdr-chdr=%d, set=0x%lx convert_utf8=%d\n", chdr, xhdr, *hdr, hdrlist->dd_len, xhdr-chdr, (unsigned long)set, convert_utf8);

    if (set) {
      if (!(set & hdrlist->inset))
	return NULL;
    } else {
      if (hdrlist->dd_len!=xhdr-chdr)
	return NULL;
      if (cmp(hdrlist->header, chdr, xhdr-chdr))
	return NULL;
    }
  }
  if (convert_utf8) {
    ret = utf8_header(hdrlist);
    if (ret == NULL)
      return NULL; /* TODO: check memory failure (errno == ENOMEM) */
    ret += hdrlist->dd_len;
  } else {
    ret = hdrlist->header+hdrlist->dd_len;
  }
  if (*ret==':') ret++;
  while (*ret == ' ' || *ret == '\t') ret++;
  DEBUG("ret=%s\n", ret);
  return ret;
}

/* 
 * check RE in hdr headers from hdrlist
 */

int regchk(RE *re, char *hdr, int sense,
	   struct headers *hdrlist, int type) {
  int ret;
  int bang  = FALSE;

  if (re->err)
    return 0;

  if (*hdr=='!') {
    bang  = TRUE;
    hdr++;
  }

  DEBUG("hdr=%s, sense=%d bang=%d\n", hdr, sense, bang);

  while(hdrlist) {
    char *xhdr=hdr;
    DEBUG("hdrlist=%.*s, type=%d, hdrlist->type=%d, hdrlist->inset=%x\n",
	  hdrlist->dd_len, hdrlist->header, 
	  type, hdrlist->type, hdrlist->inset);
    if (hdrlist->type <= type) {
      while(xhdr) {
	char *h;
	if (*xhdr=='*') {
	  xhdr=NULL;
	  h = tochk(NULL, hdrlist);
	}
	else 
	  h = tochk(&xhdr, hdrlist);
	if (!h)
	  continue;
	DEBUG("check %s\n", h);
	ret=re->exec(re, h);
	if (ret==sense) {
	  DEBUG("match sense=%d bang=%d ret=%d\n", sense, bang, ret);
	  return !bang;
	}
	DEBUG("not match sense=%d bang=%d ret=%d xhdr=%s\n", sense, bang, ret, xhdr);
      }
    }
    hdrlist=hdrlist->next;
  }
  DEBUG("fallback sense=%d bang=%d\n", sense, bang);
  return bang;
}

/*
 * return -1 if mail is looping
 */

int loopchk(char *rcpt, struct headers *hdrlist) {
  while(hdrlist) {
    if (hdrlist->type == HEADER_DELIVERED_TO) {
      char *s = hdrlist->header + sizeof("Delivered-To");
      while(isspace((int)*s)) s++;
      if (strcasecmp(s, rcpt) == 0)
	return -1;
    }
    hdrlist=hdrlist->next;
  } 
  return 0;
}

/*
 * check len bytes in body
 */

int bodchk(struct fstate *state, RE *re, int len) {
  char *buf;
  int buflen;
  int start;
  int ret;
  if (state->bodycheck_maxlen<0 || len == 0) {
    return 0;
  }
  if (len<0) {
    buflen = -len;
    if (state->bodycheck_maxlen> 0 && buflen > state->bodycheck_maxlen)
      return 0;

    if ( buflen > state->bodysize) {
      start = 0;
      len = buflen;
    } else {
      start = state->bodysize - buflen;
    }
    
  } else {
    start = 0;
    buflen = len;
    if (state->bodycheck_maxlen> 0 && buflen > state->bodycheck_maxlen)
      return 0;
    if (buflen > state->bodysize) {
      buflen = state->bodysize;
    }
  }
  if ((buf = malloc(buflen+1))==NULL) {
    MemoryError(state);
    return 0;
  }
#if 0
  fseek(state->spool, start, SEEK_SET);
  readlen = fread(buf, 1, buflen, state->spool);
#endif

  memcpy(buf, state->body + start, buflen);
  buf[buflen]=0;

  /*
  printf("start=%d, buflen=%d, len=%d, readlen=%d, buf=%s", start, buflen, len, readlen, buf);
  */

  ret=re->exec(re, buf);
  free(buf);
  return !ret;
}

char *mod_run(struct fstate *state, char *module) {
  struct slist_elt *elt;
  char *message;

  if (state->in_module)
    return NULL;
  if (*module==0 || strchr(module, '/') || *module=='.')
    return NULL;
  if (strlen(state->module_path)+strlen(module)>=PATH_MAX-3)
    return NULL;
  if ((elt = slist_search(&state->mod_cache, module)) == NULL) {
    char file[PATH_MAX];
    long oldscore;
    extern struct cmd mod_cmdlist[];

    if (state->mod_message) {
      free(state->mod_message);
      state->mod_message=NULL;
    }
    oldscore = state->score;
    state->score=0;
    state->tag[0]=0;
    strcpy(file, state->module_path);
    strcat(file, "/");
    strcat(file, module);
    state->in_module++;
    filterfile_cmd(state, file, mod_cmdlist, NULL);
    state->in_module--;
    message = state->mod_message;
    state->mod_message=NULL;
    if (message == NULL) {
      message = "";
    }
    slist_put(&state->mod_cache, module, message);
    state->score = oldscore;
    state->tag[0]=0;
  } else {
    message = elt->value;
  }
  return message;
}

int modchk(struct fstate *state, RE *re, char *module, int sense) {
  char *message;
  int ret;
  if (state->in_module)
    return FALSE;
  message = mod_run(state, module);
  DEBUG("module %s: %s\n", module, message);
  if (message) {
    ret = re->exec(re, message);
    DEBUG("match %s: %d (sense = %d)\n", message, ret, sense);
    return ret == sense ? TRUE : FALSE;
  } else {
    return FALSE;
  }
}

==> lmtpd-0.9.24.tar.gz: lmtpd-0.9.24/chk.c <==
/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: chk.c,v 1.30 2004/10/26 09:59:19 lwa Exp $";



#include <sys/types.h>
#include <regex.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "lmtp.h"

#define TRUE  1
#define FALSE 0

#if 0
#include <stdio.h> 
#define DEBUG(x...) printf(##x)
#else
#define DEBUG(x...)
#endif

static char *utf8_header(struct headers *h) {
  if (h->utf8)
    return h->utf8;
  switch(h->type) {
  case HEADER_DELIVERED_TO:
  case HEADER_RETURN_PATH:
  case HEADER_MESSAGE_ID:
  case HEADER_RECEIVED:
    h->utf8 = h->header;
  default:
    h->utf8 = rfc2047_read(h->header, h->header+h->dd_len);
  }
  return h->utf8;
}

/*
 * return 0 if RE is valid
 */

int valid_re(RE *re) {
  switch(re->err) {
  case RE_ERR_MEMORY:
    errno = ENOMEM;
    return -1;
    break;
  case RE_ERR_SYNTAX:
    errno = EINVAL;
    return -1;
    break;
  }
  return 0;
}

/*
 *  extract first hdr header from hdrlist
 */

char *tochk(char **hdr, struct headers *hdrlist) {
  char *ret;
  int convert_utf8 = 0;

  int (*cmp)(const char *, const char *, size_t);

  if (hdr) {
    char *xhdr=*hdr;
    char *chdr=*hdr;
    uint32_t set = 0;
    convert_utf8 = 0;

    if (*xhdr=='\\') {
      xhdr++; chdr++;
      cmp = strncmp;
    } else {
      if (*xhdr == '~') {
	xhdr++;
	chdr++;
	convert_utf8 = 1;
      }
      if (*xhdr == '%')
	set = get_set(xhdr[1]);
      cmp = strncasecmp;
    }
    while(1) {
      if (*xhdr == '\0' || *xhdr == ' ' || *xhdr == '\t') {
	*hdr=NULL;
	return NULL;
      }
      if (*xhdr==':')
	break;
      xhdr++;
    }
    *hdr=xhdr;
    while (**hdr==':') (*hdr)++;

    DEBUG("chdr=%s, xhdr=%s, *hdr=%s, dd_len=%d, xhdr-chdr=%d, set=0x%lx convert_utf8=%d\n", chdr, xhdr, *hdr, hdrlist->dd_len, xhdr-chdr, (unsigned long)set, convert_utf8);

    if (set) {
      if (!(set & hdrlist->inset))
	return NULL;
    } else {
      if (hdrlist->dd_len!=xhdr-chdr)
	return NULL;
      if (cmp(hdrlist->header, chdr, xhdr-chdr))
	return NULL;
    }
  }
  if (convert_utf8) {
    ret = utf8_header(hdrlist);
    if (ret == NULL)
      return NULL; /* TODO: check memory failure (errno == ENOMEM) */
    ret += hdrlist->dd_len;
  } else {
    ret = hdrlist->header+hdrlist->dd_len;
  }
  if (*ret==':') ret++;
  while (*ret == ' ' || *ret == '\t') ret++;
  DEBUG("ret=%s\n", ret);
  return ret;
}

/* 
 * check RE in hdr headers from hdrlist
 */

int regchk(RE *re, char *hdr, int sense,
	   struct headers *hdrlist, int type) {
  int ret;
  int bang  = FALSE;

  if (re->err)
    return 0;

  if (*hdr=='!') {
    bang  = TRUE;
    hdr++;
  }

  DEBUG("hdr=%s, sense=%d bang=%d\n", hdr, sense, bang);

  while(hdrlist) {
    char *xhdr=hdr;
    DEBUG("hdrlist=%.*s, type=%d, hdrlist->type=%d, hdrlist->inset=%x\n",
	  hdrlist->dd_len, hdrlist->header, 
	  type, hdrlist->type, hdrlist->inset);
    if (hdrlist->type <= type) {
      while(xhdr) {
	char *h;
	if (*xhdr=='*') {
	  xhdr=NULL;
	  h = tochk(NULL, hdrlist);
	}
	else 
	  h = tochk(&xhdr, hdrlist);
	if (!h)
	  continue;
	DEBUG("check %s\n", h);
	ret=re->exec(re, h);
	if (ret==sense) {
	  DEBUG("match sense=%d bang=%d ret=%d\n", sense, bang, ret);
	  return !bang;
	}
	DEBUG("not match sense=%d bang=%d ret=%d xhdr=%s\n", sense, bang, ret, xhdr);
      }
    }
    hdrlist=hdrlist->next;
  }
  DEBUG("fallback sense=%d bang=%d\n", sense, bang);
  return bang;
}

/*
 * return -1 if mail is looping
 */

int loopchk(char *rcpt, struct headers *hdrlist) {
  while(hdrlist) {
    if (hdrlist->type == HEADER_DELIVERED_TO) {
      char *s = hdrlist->header + sizeof("Delivered-To");
      while(isspace((int)*s)) s++;
      if (strcasecmp(s, rcpt) == 0)
	return -1;
    }
    hdrlist=hdrlist->next;
  } 
  return 0;
}

/*
 * check len bytes in body
 */

int bodchk(struct fstate *state, RE *re, int len) {
  char *buf;
  int buflen;
  int start;
  int ret;
  if (state->bodycheck_maxlen<0 || len == 0) {
    return 0;
  }
  if (len<0) {
    buflen = -len;
    if (state->bodycheck_maxlen> 0 && buflen > state->bodycheck_maxlen)
      return 0;

    if ( buflen > state->bodysize) {
      start = 0;
      len = buflen;
    } else {
      start = state->bodysize - buflen;
    }
    
  } else {
    start = 0;
    buflen = len;
    if (state->bodycheck_maxlen> 0 && buflen > state->bodycheck_maxlen)
      return 0;
    if (buflen > state->bodysize) {
      buflen = state->bodysize;
    }
  }
  if ((buf = malloc(buflen+1))==NULL) {
    MemoryError(state);
    return 0;
  }
#if 0
  fseek(state->spool, start, SEEK_SET);
  readlen = fread(buf, 1, buflen, state->spool);
#endif

  memcpy(buf, state->body + start, buflen);
  buf[buflen]=0;

  /*
  printf("start=%d, buflen=%d, len=%d, readlen=%d, buf=%s", start, buflen, len, readlen, buf);
  */

  ret=re->exec(re, buf);
  free(buf);
  return !ret;
}

char *mod_run(struct fstate *state, char *module) {
  struct slist_elt *elt;
  char *message;

  if (state->in_module)
    return NULL;
  if (*module==0 || strchr(module, '/') || *module=='.')
    return NULL;
  if (strlen(state->module_path)+strlen(module)>=PATH_MAX-3)
    return NULL;
  if ((elt = slist_search(&state->mod_cache, module)) == NULL) {
    char file[PATH_MAX];
    long oldscore;
    extern struct cmd mod_cmdlist[];

    if (state->mod_message) {
      free(state->mod_message);
      state->mod_message=NULL;
    }
    oldscore = state->score;
    state->score=0;
    state->tag[0]=0;
    strcpy(file, state->module_path);
    strcat(file, "/");
    strcat(file, module);
    state->in_module++;
    filterfile_cmd(state, file, mod_cmdlist, NULL);
    state->in_module--;
    message = state->mod_message;
    state->mod_message=NULL;
    if (message == NULL) {
      message = "";
    }
    slist_put(&state->mod_cache, module, message);
    state->score = oldscore;
    state->tag[0]=0;
  } else {
    message = elt->value;
  }
  return message;
}

int modchk(struct fstate *state, RE *re, char *module, int sense) {
  char *message;
  int ret;
  if (state->in_module)
    return FALSE;
  message = mod_run(state, module);
  DEBUG("module %s: %s\n", module, message);
  if (message) {
    ret = re->exec(re, message);
    DEBUG("match %s: %d (sense = %d)\n", message, ret, sense);
    return ret == sense ? TRUE : FALSE;
  } else {
    return FALSE;
  }
}
